package edu.kit.ipd.descartes.adaptation.model.dmm.util;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

import java.util.List;

import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.junit.Before;
import org.junit.Test;

import edu.kit.ipd.descartes.adaptation.AbstractTest;
import edu.kit.ipd.descartes.adaptation.model.dmm.util.DmmModelActionHelper;
import edu.kit.ipd.descartes.adaptation.model.dmm.util.DmmModelChanger;
import edu.kit.ipd.descartes.core.AdaptableEntity;
import edu.kit.ipd.descartes.mm.adaptation.AdaptationScope;
import edu.kit.ipd.descartes.mm.containerrepository.ContainerRepository;
import edu.kit.ipd.descartes.mm.containerrepository.ContainerTemplate;
import edu.kit.ipd.descartes.mm.containerrepository.ContainerrepositoryFactory;
import edu.kit.ipd.descartes.mm.containerrepository.ContainerrepositoryPackage;
import edu.kit.ipd.descartes.mm.resourceconfiguration.ActiveResourceSpecification;
import edu.kit.ipd.descartes.mm.resourceconfiguration.NumberOfParallelProcessingUnits;
import edu.kit.ipd.descartes.mm.resourceconfiguration.PassiveResourceCapacity;
import edu.kit.ipd.descartes.mm.resourceconfiguration.PassiveResourceSpecification;
import edu.kit.ipd.descartes.mm.resourceconfiguration.ProcessingResourceSpecification;
import edu.kit.ipd.descartes.mm.resourceconfiguration.ResourceconfigurationFactory;
import edu.kit.ipd.descartes.mm.resourcelandscape.ComputingInfrastructure;
import edu.kit.ipd.descartes.mm.resourcelandscape.Container;
import edu.kit.ipd.descartes.mm.resourcelandscape.DataCenter;
import edu.kit.ipd.descartes.mm.resourcelandscape.DistributedDataCenter;
import edu.kit.ipd.descartes.mm.resourcelandscape.ResourcelandscapeFactory;
import edu.kit.ipd.descartes.mm.resourcelandscape.ResourcelandscapePackage;
import edu.kit.ipd.descartes.mm.resourcelandscape.RuntimeEnvironment;

public class DmmModelActionHelperTest extends AbstractTest {

    private static final double MAX_VALUE = 4.0d;
    private static final String CANDIDATE = "Candidate";
    private static final double MIN_VALUE = 2.0d;
    AdaptableEntity aEntity = null;
    AdaptableEntity aEntity2 = null;
    RuntimeEnvironment vm = ResourcelandscapeFactory.eINSTANCE.createRuntimeEnvironment();
    ComputingInfrastructure ci = ResourcelandscapeFactory.eINSTANCE.createComputingInfrastructure();
    ContainerTemplate template = ContainerrepositoryFactory.eINSTANCE.createContainerTemplate();
    ActiveResourceSpecification aSpec = ResourceconfigurationFactory.eINSTANCE.createActiveResourceSpecification();;
    PassiveResourceSpecification pSpec = ResourceconfigurationFactory.eINSTANCE.createPassiveResourceSpecification();;
    
    @Before
    public void setUp() throws Exception {
        super.setUp();
        NumberOfParallelProcessingUnits parProcUnits = ResourceconfigurationFactory.eINSTANCE.createNumberOfParallelProcessingUnits();
        ProcessingResourceSpecification cpu = ResourceconfigurationFactory.eINSTANCE.createProcessingResourceSpecification();
        cpu.setNrOfParProcUnits(parProcUnits);
        PassiveResourceCapacity passiveCapacity = ResourceconfigurationFactory.eINSTANCE.createPassiveResourceCapacity();

        pSpec.setCapacity(passiveCapacity);
        aSpec.getProcessingResourceSpecifications().add(cpu);
        
        vm.getConfigSpec().add(aSpec);
        aEntity = parProcUnits;
        
        template.getTemplateConfig().add(pSpec);
        aEntity2 = passiveCapacity;
    }

    @Test
    public void testFindParentConfiguredInstance() {
        EObject foundClass = null;
        
        foundClass = DmmModelActionHelper.findParentConfiguredInstance(aEntity);
        assertEquals(ResourcelandscapePackage.RUNTIME_ENVIRONMENT, foundClass.eClass().getClassifierID());
        
        ci.getConfigSpec().add(aSpec);
        foundClass = DmmModelActionHelper.findParentConfiguredInstance(aEntity);
        assertEquals(ResourcelandscapePackage.COMPUTING_INFRASTRUCTURE, foundClass.eClass().getClassifierID());
        
        foundClass = DmmModelActionHelper.findParentConfiguredInstance(aEntity2);
        assertEquals(ContainerrepositoryPackage.CONTAINER_TEMPLATE, foundClass.eClass().getClassifierID());
    }
    
    @Test
    public void testFilterEntities() {
        Container c1 = ResourcelandscapeFactory.eINSTANCE.createComputingInfrastructure();
        Container c2 = ResourcelandscapeFactory.eINSTANCE.createComputingInfrastructure();
        
        
        RuntimeEnvironment re1 = ResourcelandscapeFactory.eINSTANCE.createRuntimeEnvironment();
        re1.setName(CANDIDATE + DmmModelChanger.COPY_MARK);
        re1.setId(EcoreUtil.generateUUID());
        re1.getConfigSpec().add(aSpec);
        
        RuntimeEnvironment re2 = EcoreUtil.copy(re1);
        re2.setId(EcoreUtil.generateUUID());
        RuntimeEnvironment re3 = EcoreUtil.copy(re1);
        re3.setId(EcoreUtil.generateUUID());
        re3.setTemplate(template);
        
        re1.setContainedIn(c1);
        re2.setContainedIn(c1);
        re3.setContainedIn(c2);
        
        EList<RuntimeEnvironment> entityList = new BasicEList<RuntimeEnvironment>();
        entityList.add(re1);
        entityList.add(re2);
        entityList.add(re3);
        
        RuntimeEnvironment comparator = ResourcelandscapeFactory.eINSTANCE.createRuntimeEnvironment();
        comparator.setName(CANDIDATE);
        comparator.setId(EcoreUtil.generateUUID());
        comparator.getConfigSpec().add(aSpec);
        comparator.setTemplate(template);
        comparator.setContainedIn(c2);
        
        RuntimeEnvironment result = DmmModelActionHelper.filterCopiedRuntimeEnvironment(entityList, comparator);
        assertEquals(re3, result);
        
        entityList.remove(re3);
        assertEquals(null, DmmModelActionHelper.filterCopiedRuntimeEnvironment(entityList, comparator));
    }

    @Test
    public void testQueryUnfilledRuntimeEnvironment() {
        DistributedDataCenter testDdc = ResourcelandscapeFactory.eINSTANCE.createDistributedDataCenter();
        DataCenter dc1 = ResourcelandscapeFactory.eINSTANCE.createDataCenter();
        testDdc.getConsistsOf().add(dc1);
        ComputingInfrastructure ci1 = ResourcelandscapeFactory.eINSTANCE.createComputingInfrastructure();
        dc1.getContains().add(ci1);
        ContainerTemplate template = ContainerrepositoryFactory.eINSTANCE.createContainerTemplate();
        RuntimeEnvironment re1 = ResourcelandscapeFactory.eINSTANCE.createRuntimeEnvironment();
        RuntimeEnvironment re2 = ResourcelandscapeFactory.eINSTANCE.createRuntimeEnvironment();
        ci1.getContains().add(re1);
        ci1.getContains().add(re2);
        re1.setTemplate(template);
        re2.setTemplate(template);
        re1.setName("RE1");
        re2.setName("RE2");
        
        ActiveResourceSpecification active1 = ResourceconfigurationFactory.eINSTANCE.createActiveResourceSpecification();
        re1.getConfigSpec().add(active1);
        ActiveResourceSpecification active2 = ResourceconfigurationFactory.eINSTANCE.createActiveResourceSpecification();
        re2.getConfigSpec().add(active2);
        
        ProcessingResourceSpecification procSpec1 = ResourceconfigurationFactory.eINSTANCE.createProcessingResourceSpecification();
        NumberOfParallelProcessingUnits nr1 = ResourceconfigurationFactory.eINSTANCE.createNumberOfParallelProcessingUnits();
        nr1.setNumber(4);
        procSpec1.setNrOfParProcUnits(nr1);
        active1.getProcessingResourceSpecifications().add(procSpec1);
        
        ProcessingResourceSpecification procSpec2 = ResourceconfigurationFactory.eINSTANCE.createProcessingResourceSpecification();
        NumberOfParallelProcessingUnits nr2 = ResourceconfigurationFactory.eINSTANCE.createNumberOfParallelProcessingUnits();
        nr2.setNumber(2);
        procSpec2.setNrOfParProcUnits(nr2);
        active2.getProcessingResourceSpecifications().add(procSpec2);
        
        RuntimeEnvironment res = DmmModelActionHelper.queryUnfilledRuntimeEnvironment(re1, MAX_VALUE);
        assertTrue(EcoreUtil.equals(re2, res));
    }
    
    @Test
    public void testQueryNotEmptyRuntimeEnvironment() {
        DistributedDataCenter testDdc = ResourcelandscapeFactory.eINSTANCE.createDistributedDataCenter();
        DataCenter dc1 = ResourcelandscapeFactory.eINSTANCE.createDataCenter();
        testDdc.getConsistsOf().add(dc1);
        ComputingInfrastructure ci1 = ResourcelandscapeFactory.eINSTANCE.createComputingInfrastructure();
        dc1.getContains().add(ci1);
        ContainerTemplate template = ContainerrepositoryFactory.eINSTANCE.createContainerTemplate();
        RuntimeEnvironment re1 = ResourcelandscapeFactory.eINSTANCE.createRuntimeEnvironment();
        RuntimeEnvironment re2 = ResourcelandscapeFactory.eINSTANCE.createRuntimeEnvironment();
        RuntimeEnvironment re3 = ResourcelandscapeFactory.eINSTANCE.createRuntimeEnvironment();
        ci1.getContains().add(re1);
        ci1.getContains().add(re2);
        ci1.getContains().add(re3);
        re1.setTemplate(template);
        re2.setTemplate(template);
        re3.setTemplate(template);
        re1.setName("RE1");
        re2.setName("RE2");
        re3.setName("RE3-NO_CONFIG");
        
        ActiveResourceSpecification active1 = ResourceconfigurationFactory.eINSTANCE.createActiveResourceSpecification();
        re1.getConfigSpec().add(active1);
        ActiveResourceSpecification active2 = ResourceconfigurationFactory.eINSTANCE.createActiveResourceSpecification();
        re2.getConfigSpec().add(active2);
        
        ProcessingResourceSpecification procSpec1 = ResourceconfigurationFactory.eINSTANCE.createProcessingResourceSpecification();
        NumberOfParallelProcessingUnits nr1 = ResourceconfigurationFactory.eINSTANCE.createNumberOfParallelProcessingUnits();
        nr1.setNumber(4);
        procSpec1.setNrOfParProcUnits(nr1);
        active1.getProcessingResourceSpecifications().add(procSpec1);
        
        ProcessingResourceSpecification procSpec2 = ResourceconfigurationFactory.eINSTANCE.createProcessingResourceSpecification();
        NumberOfParallelProcessingUnits nr2 = ResourceconfigurationFactory.eINSTANCE.createNumberOfParallelProcessingUnits();
        nr2.setNumber(2);
        procSpec2.setNrOfParProcUnits(nr2);
        active2.getProcessingResourceSpecifications().add(procSpec2);
        
        RuntimeEnvironment res = DmmModelActionHelper.queryNotEmptyRuntimeEnvironment(re1, MIN_VALUE);
        assertTrue(EcoreUtil.equals(re1, res));
        
        ci1.getContains().remove(re1);
        res = DmmModelActionHelper.queryNotEmptyRuntimeEnvironment(re2, MIN_VALUE);
        assertTrue(res == null);
    }
    
    @Test 
    public void testQueryContainersWithSimilarSpecs() {
        DistributedDataCenter testDdc = ResourcelandscapeFactory.eINSTANCE.createDistributedDataCenter();
        DataCenter dc1 = ResourcelandscapeFactory.eINSTANCE.createDataCenter();
        testDdc.getConsistsOf().add(dc1);
        ComputingInfrastructure ci1 = ResourcelandscapeFactory.eINSTANCE.createComputingInfrastructure();
        ComputingInfrastructure ci2 = ResourcelandscapeFactory.eINSTANCE.createComputingInfrastructure();
        dc1.getContains().add(ci1);
        dc1.getContains().add(ci2);
        ContainerTemplate template = ContainerrepositoryFactory.eINSTANCE.createContainerTemplate();
        RuntimeEnvironment re1 = ResourcelandscapeFactory.eINSTANCE.createRuntimeEnvironment();
        RuntimeEnvironment re2 = ResourcelandscapeFactory.eINSTANCE.createRuntimeEnvironment();
        RuntimeEnvironment re3 = ResourcelandscapeFactory.eINSTANCE.createRuntimeEnvironment();
        ci1.getContains().add(re1);
        ci1.getContains().add(re2);
        ci2.getContains().add(re3);
        re1.setTemplate(template);
        re2.setTemplate(template);
        re3.setTemplate(template);
        re1.setName("RE1");
        re2.setName("RE2");
        re3.setName("RE3" + DmmModelChanger.COPY_MARK);
       
        ActiveResourceSpecification active1 = ResourceconfigurationFactory.eINSTANCE.createActiveResourceSpecification();
        ActiveResourceSpecification active2 = ResourceconfigurationFactory.eINSTANCE.createActiveResourceSpecification();
        
        ProcessingResourceSpecification procSpec1 = ResourceconfigurationFactory.eINSTANCE.createProcessingResourceSpecification();
        NumberOfParallelProcessingUnits nr1 = ResourceconfigurationFactory.eINSTANCE.createNumberOfParallelProcessingUnits();
        nr1.setNumber(4);
        procSpec1.setNrOfParProcUnits(nr1);
        active1.getProcessingResourceSpecifications().add(procSpec1);
        
        ProcessingResourceSpecification procSpec2 = ResourceconfigurationFactory.eINSTANCE.createProcessingResourceSpecification();
        NumberOfParallelProcessingUnits nr2 = ResourceconfigurationFactory.eINSTANCE.createNumberOfParallelProcessingUnits();
        nr2.setNumber(2);
        procSpec2.setNrOfParProcUnits(nr2);
        active2.getProcessingResourceSpecifications().add(procSpec2);

        re1.getConfigSpec().add(active1);
        re2.getConfigSpec().add(active2);
        re3.getConfigSpec().add(EcoreUtil.copy(active1));
        
        List<Container> result = DmmModelActionHelper.queryContainersWithSimilarSpecs(re1);
        assertEquals(1, result.size());
        assertTrue( ! result.contains(re2));
    }
    
    @Test
    public void testFindSuitableContainer() {
        ContainerTemplate ct = ContainerrepositoryFactory.eINSTANCE.createContainerTemplate();
        ct.setId("TEST_ID");
        ct.setName("CONTAINER_TEMPLATE");
        
        RuntimeEnvironment c1 = ResourcelandscapeFactory.eINSTANCE.createRuntimeEnvironment();
        c1.setId("C1-ID");
        c1.setName("C1");
        c1.setTemplate(ct);
        RuntimeEnvironment c2 = ResourcelandscapeFactory.eINSTANCE.createRuntimeEnvironment();
        c2.setId("C2-ID");
        c2.setName("Bad container C2");
        RuntimeEnvironment c3  = ResourcelandscapeFactory.eINSTANCE.createRuntimeEnvironment();
        c3.setId("C3-ID");
        c3.setName("C3");
        c3.setTemplate(ct);
        
        ct.getReferringContainers().add(c1);
        ct.getReferringContainers().add(c3);
        
        Container result = DmmModelActionHelper.findSuitableContainer(ct, AdaptationScope.RANDOM);
        assertTrue(EcoreUtil.equals(result, c3) || EcoreUtil.equals(result, c1));
        assertTrue(!EcoreUtil.equals(result, c2));
    }

}
